module AST2IR where

import qualified OP
import qualified AST
import qualified IR

import Temp

forVal :: TMonad m => AST.Exp -> m IR.Exp
forVal (AST.Const i) = return (IR.CONST i)
forVal (AST.Var i) = return (IR.MEM (IR.CONST i))
forVal (AST.Binop op e1 e2) = do
       ir1 <- forVal e1
       ir2 <- forVal e2
       return (IR.BINOP op ir1 ir2)
forVal (AST.Relop op e1 e2) = do
       ltrue <- newLabel
       lfalse <- newLabel
       lafter <- newLabel
       tresult <- newTemp
       e1' <- forVal e1
       e2' <- forVal e2
       return $
           IR.ESEQ (IR.CJUMP op e1' e2' ltrue lfalse :
                   [IR.LABEL ltrue, IR.MOVE (IR.TEMP tresult) (IR.CONST 1), IR.JUMP lafter] ++
                   [IR.LABEL lfalse, IR.MOVE (IR.TEMP tresult) (IR.CONST 0), IR.JUMP lafter] ++
                   [IR.LABEL lafter])
                   (IR.TEMP tresult)

forCond :: TMonad m => AST.Exp -> IR.Label -> IR.Label -> m [IR.Stm]
forCond (AST.Const i) lt lf | i /= 0 = return [IR.JUMP lt]
forCond (AST.Const i) lt lf | i == 0 = return [IR.JUMP lf]
forCond (AST.Relop op e1 e2) lt lf = do
    e1' <- forVal e1
    e2' <- forVal e2
    return [IR.CJUMP op e1' e2' lt lf]
forCond e lt lf =
    forVal e >>= \e' -> return [IR.CJUMP OP.Equ e' (IR.CONST 0) lf lt]

forEff :: TMonad m => AST.Stm -> m [IR.Stm]
forEff (AST.EXP e) = forVal e >>= \e' -> return [IR.EXP e']
forEff (AST.Assign i e) = do
    e' <- forVal e
    return [IR.MOVE (IR.MEM (IR.CONST i)) e']
forEff (AST.SEQ s1 s2) = forEff s1 >>= \s1' -> forEff s2 >>= \s2' -> return (s1' ++ s2')
forEff (AST.If e s1 s2) = do
    ltrue <- newLabel
    lfalse <- newLabel
    lafter <- newLabel
    s0' <- forCond e ltrue lfalse
    s1' <- forEff s1
    s2' <- forEff s2
    return (s0' ++
            [IR.LABEL ltrue] ++ s1' ++ [IR.JUMP lafter] ++
            [IR.LABEL lfalse] ++ s2' ++ [IR.JUMP lafter] ++
            [IR.LABEL lafter])
forEff (AST.While e s1) = do
    ltrue <- newLabel
    lfalse <- newLabel
    ltest <- newLabel
    s0' <- forCond e ltrue lfalse
    s1' <- forEff s1
    return (IR.JUMP ltest :
            [IR.LABEL ltest] ++ s0' ++
            [IR.LABEL ltrue] ++ s1' ++ [IR.JUMP ltest] ++
            [IR.LABEL lfalse])

forProgram s = do
    lentry <- newLabel
    lexit <- newLabel
    s' <- forEff s
    return (IR.LABEL lentry : s' ++ [IR.JUMP lexit])

